"""
CoA Creation | Cannlytics
Created on Sat Feb  6 12:48:29 2021
Author: Keegan Skeate <keegan@cannlytics.com>
"""

# Allow users to create CoA's with a myriad of templates!

# Import any docx template styed with Django formatting and let'r'rip.

# Use your own templates or download the started templates and customize them
# to your heart's galore.


def create_coa():
    """
    Creates a certificate of analysis.
    """


def merge_coas():
    """
    Merge two CoAs together, either existing templates
    or existing PDFs.
    """